/**
 * Preloads images, storing them in a public index object.
 * Images are passed into this object in the src array, it preloads those images
 * storing them in this.index, local filename. If this sample src object was
 * passed into the object:
 *		['img/image.png', 'img/image2.png', fakepath/img/image3.gif]
 * Then the images would be accesible as:
 *		this.index['image'], this.index['image2'], this.index['image3']
 * When all images have been loaded the onload function is called. By default
 * this function is set to an empty function.
 * @param {array} src - list of image sources to be loaded
 * @return {this}
 */
function ImgCache( src ) {


	// *** INIT *** //

	var imgCache = this;

	this.index  = {};
	this.toLoad = src.length;
	this.loaded = 0;
	this.ready  = false;
	this.onload = function(){};

	for( var i = 0; i < src.length; i++ )
		loadImage( src[ i ]);

	// *** METHODS *** //

	/**
	 * Tracks progress of image loading. Calls this.onload when completed.
	 */
	function loadProgress() {

		if(++imgCache.loaded === imgCache.toLoad)
			imgCache.onload();

	};

	/**
	 * Loads and stores image in index[ imagename ]
	 * @param {String} imageSource - source of image to load
	 */
	function loadImage( imageSource ){

		var imageName = parseFilename( imageSource );

		imgCache.index[ imageName ] = new Image();
		imgCache.index[ imageName ].src = imageSource;
		imgCache.index[ imageName ].onload = loadProgress;

	};

	/**
	 * Get filename without file extension
	 * @param {String} src - source of which to parse name
	 * @return {String} - filename
	 */
	function parseFilename( src ){

		var filename = '',
			firstChar = src.lastIndexOf("/") + 1,
			length = src.length;

		for( var char = firstChar; char < length; char++ ){
			if( src[ char ] === "." ) break;
			filename += src[ char ];
		}

		return filename;

	}

	return this;

};


/**
 * Manipulates and draws images to canvas element.
 * Object requires a 2d context from the canvas to which will be drawn, and an
 * image object or list of frames of an animation. Using a public draw function
 * the image is manipulated by rotation, size, position etc. More info on
 * this.draw comment doc.
 * @param {CanvasRenderingContext2D} c : context in which to draw image
 * @param {(Image | Image[])} image : Image object or array of images to draw
 */
function Img( c, image ){

	// *** INIT *** //

    this.img = [];

	var x, y, w, h, xOrigin, yOrigin,
		parent        = this,
		frame         = 0,
		isAnimated    = false,
		rotation      = 0,
		rotationSpeed = 0;

	init();


	// *** METHODS *** //

	// public

	/**
	 * Draws image to screen, image can be modified with the param object that
	 * is passed in.
	 * @param {Object} param - parameter object.
	 * @param {Number} param.x [x=0] - x coordinate to draw image.
	 * @param {Number} param.y [x=0] - y coordinate to draw image.
	 * @param {Number} param.w [w=image.width] - width to draw image.
	 * @param {Number} param.h [h=image.height] - height to draw image.
	 * @param {Number} param.scale [scale=1] - scale both param.w and param.h.
	 		ex: if(w = 100 && scale = 2) drawing width = 200
	 * @param {Number} param.xScale [xScale=1] - scale width of image.
	 * @param {Number} param.yScale [yScale=1] - scale height of image.
	 * @param {Number} param.xOrigin [xOrigin=0] - x coordinate in which to draw
	 	and rotate sprite.
	 * @param {Number} param.yOrigin [yOrigin=0] - y coordinate in which to draw
	 	and rotate sprite.
	 * @param {Boolean} param.centerOrigin [centerOrigin=false] - whether to set
	 	the origin to the center of image.
	 * @param {Boolean} param.animate [animate=false] - whether to animate
	 	sequence of images.
	 * @param {Number} param.animationSpeed [animationSpeed=1] - speed to switch
	  	images, if animation speed is set greater than one some images will be
		cut from the animation, if animation speed is below 1 the animation
		frame will not change every frame, values can be decimal values. For
		example if animationSpeed is equal to .5, the image frame will change
		every second time this function is called.
	 * @param {Number} param.f [f=0] - animation frame to draw, this parameter
	 	can only be used if the images are not being animated.
	 * @param {Number} param.rotationSpd [rotationSpd=0] - rotation that occurs
	  	on image each frame. Positive values rotate clockwise and negative
		values cause image to rotate counter clockwise.
	 * @param {Number} param.roation [rotation=0] - rotation of image to draw.
	  	If this parameter is defined it will override the rotationSpd value and
		the image will have a static rotation.
	 * @param {Number} param.opacity [opacity=1] - opacity in which to draw
	 	image.
	 * @param {color} param.shadowColor [shadowColor='#000'] - value to pass to
	 	the contexts shadowColor property.
	 * @param {Number} param.shadowBlur [shadowBlur=0] - value to pass to the
	 	contexts shadowBlur property.
	 * @param {Number} param.shadowOffsetX [shadowOffsetX=0] - value to pass to
   	 	the contexts shadowOffsetX property.
	 * @param {Number} param.shadowOffsetY [shadowOffsetY=0] - value to pass to
  	 	the contexts shadowOffsetY property.
	 */
    this.draw = function( param ){

		var param = param || {};

		getCoordinates(param);
		getDimensions(param);
		getOrigin(param);
		animate(param);
		rotate(param);

		drawSprite(param);

    };


	// private

	function init(){

		// single image
		if(!Array.isArray(image)) parent.img[0] = image;

		else{ // array of images
			for(var i=0; i<image.length; i++)
				parent.img[i] = image[i];
		}

	}

	function getCoordinates( param ){

		x = param.x || 0;
		y = param.y || 0;

	}

	function getDimensions( param ){

			// base dimensions
		var image      = parent.img[ Math.floor( frame )];
			baseWidth  = ( param.w != undefined ) ? param.w : image.width,
			baseHeight = ( param.h != undefined ) ? param.h : image.height,

			// scale
			scale      = ( param.scale   != undefined ) ? param.scale  : 1,
			xScale     = (( param.xScale != undefined ) ? param.xScale : 1 ) * scale,
			yScale     = (( param.yScale != undefined ) ? param.yScale : 1 ) * scale;


		w = baseWidth * xScale;
		h = baseHeight * yScale;

	}

	function getOrigin( param ){

		xOrigin = param.xOrigin || 0;
		yOrigin = param.yOrigin || 0;

		if(param.centerOrigin){
			xOrigin = w / 2;
			yOrigin = h / 2;
		}

	}

	function animate( param ){

		isAnimated = param.animate || false;

		if( !isAnimated ) frame = param.f || 0;

		else{

			var	speed   = param.animationSpeed || 1;

			frame = (frame + speed) % parent.img.length;

		}

	}

	function rotate( param ){

		rotationSpeed = param.rotateSpd || 0;
		rotation      = ( param.rotate != undefined ) ? param.rotate : rotation + rotationSpeed;

	}

	function setShadow( param ){

		c.shadowColor   = param.shadowColor   || '#000';
		c.shadowBlur    = param.shadowBlur    || 0;
		c.shadowOffsetX = param.shadowOffsetX || 0;
		c.shadowOffsetY = param.shadowOffsetY || 0;

	}

	function drawSprite( param ){

		c.save();
		setShadow( param );
		c.translate( x, y );
		c.rotate( toRadians( rotation ));
		c.globalAlpha = ( param.opacity != undefined ) ? param.opacity : 1;
		c.drawImage( parent.img[ Math.floor( frame )], -xOrigin, -yOrigin, w, h );
		c.restore();

	}

}
