/** Infinite Skies Banner Controller Object
 * @author : Jesse McCarthy
 * @id : 3270416
 * Controls all objects and interactions of the infinite skies banner.
 */


/**
 * Infinite Skies Banner Controller.
 * Object contains all banner objects such as the player star, and background
 * stars. After preloading all relavent images the animation loop begins. The
 * animation loop is called 30 times per second and only does two things. Calls
 * the this.update, and calls this.draw.
 * @param {Element<Canvas>} canvas - The canvas element that will be used as the
    Infinite Skies Banner.
 */
function InfiniteSkiesBanner( canvas ){

    // =========================== //
    // ---------- INIT ----------- //
    // =========================== //

    var controller = this,
        context    = canvas.getContext('2d'),
        mouse      = new Mouse( canvas ),
        fps        = 30,
        interval   = 1000/fps,

        // images
        sprites   = new ImgCache([
            'img/ss_banner/star.png',
            'img/ss_banner/aquaStar.png',
            'img/ss_banner/blueStar.png',
            'img/ss_banner/whiteStar.png',
            'img/ss_banner/purpleStar.png'
        ]),

        // list is used to randomize star type
        starTypes = [
            sprites.index['aquaStar'],
            sprites.index['blueStar'],
            sprites.index['whiteStar'],
            sprites.index['purpleStar']
        ],

        // banner objects
        bgStarGen = new BgStarGenerator(),
        playerStar = new PlayerStar(),
        objects = {
            bgStars    : []
        };

    // Start main loop on image load
    sprites.onload = startMainLoop;


    // ============================== //
    // ---------- METHODS ----------- //
    // ============================== //

    // *** PUBLIC METHODS *** //

    /**
     * Updates all banner objects and removes garbage.
     * This function is called in the main loop at a frequency of 30 times per
     * second.
     */
    this.update = function(){
        updateObjects();
        garbageCollection();
    };

    /**
     * Clear canvas and draw all banner objects to the canvas element.
     * This function is called in the main loop at a frequency of 30 times per
     * second.
     */
    this.draw = function(){
        context.clearRect( 0, 0, canvas.width, canvas.height );
        drawObjects();
    };


    // *** PRIVATE METHODS *** //

    // Starts main loop
    function startMainLoop(){
        setInterval( function(){
            controller.update();
            controller.draw();
        }, interval );
    }

    /**
     * Iterates through all banner objects and uses callback funciton
     * @param {requestCallback} callback : callback function to be passed into
        forEach function
     */
    function iterateObjects( callback ){
        for(var type in objects)
            objects[type].forEach( callback );
    }

    // Updates all banner objects
    function updateObjects(){

        iterateObjects( function( object ){
           object.update();
        });
        playerStar.update();
        bgStarGen.update();

    }

    // Deletes all objects whose delete attribute is equal to true.
    function garbageCollection(){

        for(var type in objects)
            objects[ type ] = objects[ type ].filter( function( object ){
                return !object.delete;
            });

    }

    // Draws all objects to canvas.
    function drawObjects(){

        iterateObjects( function( object ){
            object.draw();
        });
        playerStar.draw();

    }


    // =================================== //
    // ---------- GAME OBJECTS ----------- //
    // =================================== //

    function ObjectInterface(){
        this.delete = false;
        this.update = function(){};
        this.draw   = function(){};
    }


    /**
     * Player star class
     * Player star follows mouse if the mouse is hovering the canvas
     */
    function PlayerStar(){

        // *** INIT *** //

        ObjectInterface.call( this );

        var playerStar = this,
            sprite = new Img( context, sprites.index.star ),
            x = canvas.width / 2,
            y = canvas.height / 2,
            speed = 5;

        // *** PUBLIC METHODS *** //

        this.update = function(){
            moveTowardsMouse();
        };

        this.draw = function(){

            sprite.draw({
                x : x,
                y : y,
                w : 32,
                h : 32,
                rotateSpd : 20,
                centerOrigin : true,
                shadowColor : 'aqua',
                shadowBlur : 60
            });

        }

        // *** PRIVATE METHODS *** //

        function moveTowardsMouse(){
            // Move player towards mouse when mouse is hovering over the canvas

            if( mouse.hover ){

                if( getDistance( x, y, mouse.x, mouse.y ) > speed){

                    // follow the mouse
                    var angle = getAngle( x, y, mouse.x, mouse.y );
                    x += speed * Math.cos( angle );
                    y += speed * Math.sin( angle );

                }

                // if the stars distance is less than its speed
                else{
                    x = mouse.x;
                    y = mouse.y;
                }
            }
        }
    }


    /**
     * Background star class
     * @param x (int) : x coordinate
     * @param y (int) : y coordinate
     */
    function BgStar( x, y ){

        this.sprite = new Img( context, this.randomStarType() );

        // coordinates
        this.x = x;
        this.y = y;

        // parallax
        this.z = randomFloat( .05, 1.5 );
        this.parallaxOffset = 0;
        this.originalScrollHeight = getScrollHeight();
        this.scale = 2 - this.z;

        // animation
        this.opacity = 0;
        this.rotateSpd = randomInt( -3, 3 );
        this.fader = new Fader( 90 );

    }

    BgStar.prototype = new (function(){

        ObjectInterface.call( this );

        this.update = function(){
            this.fader.fade();
            if(this.fader.complete)
                this.delete = true;
        };

        this.draw = function(){
            this.sprite.draw({
                x : this.x,
                y : this.y - this.parallaxOffset,

                scale : this.scale,
                centerOrigin : true,
                opacity      : this.fader.opacity,
                rotateSpd    : this.rotateSpd
            });
        };

        this.randomStarType = function(){
            return starTypes[ randomInt( 0, starTypes.length - 1 )];
        };

        this.parallaxScorlling = function(){
            this.parallaxOffset = ( getScrollHeight() - this.originalScrollHeight ) / this.z;
        };

    })();

    window.addEventListener( 'scroll', function(){
        iterateObjects( function(object){ object.parallaxScorlling(); })
    });

    /**
     * Background star Generator class
     * Intermittently generates stars to a random position on the canvas
     */
    function BgStarGenerator(){
        // *** INIT *** //

        var spawnTimer = new Timer([ 3, 5 ], spawnStar );

        // *** METHODS *** //

        this.update = function(){

            if( objects.bgStars.length < 80 )
                spawnTimer.tick();

        };

        function spawnStar(){

            var x = Math.random() * canvas.width,
                y = Math.random() * canvas.height;

            objects.bgStars.push( new BgStar( x, y ));

        }
    }

}
