
// *** UTILITY FUNCTIONS *** //

Array.prototype.isEmpty = function(){
	return( this.length === 0 );
}

/**
 * Linear fade object.
 * Opacity shifts from 0.0 to 1.0 for $length amount of frames. After $length
 * amount of frames opacity shifts back from 1.0 to 0.0, in linear pattern.
 * @param {integer} length : number of frames for one half of fade in and fade
  	out cycle.
 */
function Fader( length ){

	var fader       = this,
		timeElapsed = 0,
		isFadingOut = false;

	this.complete = false;
	this.opacity = 0;

	/**
	 * Object Step where time is incremented and the opacity is shifted.
	 * @return {this}
	 */
	this.fade = function(){

		tick();
		transition();
		shiftOpacity();

		return this;

	}

	// Time incrmentation
	function tick(){

		timeElapsed = ++timeElapsed % length;

	}

	// Transitions between fading in and fading out
	function transition(){

		if( !timeElapsed ){

			if( isFadingOut ) fader.complete = true;
			else              isFadingOut    = true;

		}

	}

	// Shifts opacity value
	function shiftOpacity(){

		var fadePercent = timeElapsed / length;

		if( isFadingOut ) fader.opacity = 1 - fadePercent;
		else              fader.opacity = fadePercent;

	}

}

/**
 * Sine Cycle Object.
 * Functions much the same as the Fader object, except that it shifts in a sine
 * wave pattern with an amplitude of $amplitude in $steps amount of frames.
 * @param {integer} steps : frequency of sine wave
 * @param {number} amplitude : amplitude of sine wave
 */
function SineCycle( steps, amplitude ){

	// *** INIT *** //

	var sineCycle = this,
		step      = 0,
		amplitude = amplitude || 1;

	this.value    = 0;
	this.vertex   = false;
	this.complete = false;

	// *** METHODS *** //

	/**
	 * Sine value is computed, states are tested (vertex or completion), and
	 * time is incremented.
	 * @return {this}
	 */
	this.update = function(){

		computeValue();
		testForVertex();
		testForCompletion();
		incrementStep();

		return this;

	}

	/**
	 * Resets time, sine value, and states.
	 * @return {this}
	 */
	this.reset = function(){

		step          = 0;
		this.value    = 0;
		this.vertex   = false;
		this.complete = false;

		return this;

	};

	function computeValue(){

		sineCycle.value = amplitude * Math.sin( step / steps * Math.PI );

	}

	function testForVertex(){

		this.vertex = false;

		if( step == steps / 2 ) sineCycle.vertex = true;

	}

	function testForCompletion(){

		if( step == steps ) sineCycle.complete = true;

	}

	function incrementStep(){

		step++;

	}

}

/**
 * Executes callback function after a certain amount of steps.
 * @param {number} length - number of steps until callback is executed.
 * @param {requestCallback} callback - function to execute.
 * @param {any} args - arguments to be passed into callback function.
 */
function Timer( length, callback, args ){


	// *** INIT *** //

	var time,
		timeElapsed = 0;

		setTime();

	// *** METHODS *** //

	this.tick = function(){

		timeElapsed = ++timeElapsed % time;

		if( !timeElapsed ){
			callback( args );
			setTime();
		}

	}

	this.reset = function(){

		timeElapsed = 0;
		setTime();

	}

	function setTime(){

		if( typeof length === 'number' )       time = length;
		else if( typeof length === 'function') time = length();
		else if( Array.isArray( length ) )     time = randomInt( length[0], length[1] );

	}

}

/**
 * Tracks the mouse object relative to an element
 * @param {Element} relative - Relative element
 */
function Mouse( relative ){

	var mouse = this;
		relative = relative || document.body;

	this.x = 0; // x coordinate of mouse
	this.y = 0; // y coordinate of mouse
	this.hover = false; // whether the $relative element is being hovered

	// record mouse coordinates on mousemove
	addEventListener( 'mousemove', function( event ){

		var rect = relative.getBoundingClientRect(),
			mx   = mouse.x - rect.left,
			my   = mouse.y - rect.top;

		mouse.x = event.clientX - rect.left;
		mouse.y = event.clientY - rect.top;

	});

	// mouse enter relative element
	relative.addEventListener( 'mouseenter', function(){
		mouse.hover = true;
	});

	// mouse leave relative element
	relative.addEventListener( 'mouseleave', function(){
		mouse.hover = false;
	});

}


function getScrollHeight(){
	// CODE MODIFIED FROM
	// http://stackoverflow.com/questions/871399/cross-browser-method-for-detecting-the-scrolltop-of-the-browser-window
	// BY kennebec
	// THANK THE LORD! I SPENT TOOOOOO LONG ON THAT!

	if(typeof pageYOffset!= 'undefined'){
        //most browsers except IE before #9
        return pageYOffset;
    }
    else{
		var body = document.body, //IE 'quirks'
        	docElement = document.documentElement; //IE with doctype
		docElement = (docElement.clientHeight) ? docElement : body;
		return docElement.scrollTop;
    }
}


function isNull( variable ){

	return variable === null;

}

function limit( value, max ){

	return Math.min( value, max );

}

function within(rect){
    return (this.x > rect.x && this.x < rect.x + rect.w && this.y > rect.y && this.y < rect.y + rect.h);
}

function isDefined(v){
    return (typeof v !== 'undefined');
}

function plusOrMinus(val){
    if(randomInt(0,1))
        return val;
    return -val;
}

function randomInt( min, max ){
    return Math.floor( Math.random() * ( max - min + 1 )) + min;
}

function randomFloat( min, max ){
    return Math.random() * ( max - min ) + min;
}

function toDegrees( angle ){
    return angle * 180 / Math.PI;
}

function toRadians( angle ){
    return angle / 180 * Math.PI;
}

function getDistance(x1,y1,x2,y2){
    return Math.sqrt(Math.pow(x1-x2,2) + Math.pow(y1-y2,2));
}

//Code Modified From: https://gist.github.com/conorbuck/2606166
function getAngle( x1, y1, x2, y2 ){

	if( x1 === x2 && y1 === y2 ) return null;

	return Math.atan2( y2 - y1, x2 - x1 );

}

function drawCircle(c,x,y,size,color){
    c.beginPath();
    c.arc(x, y, size, 0, 2 * Math.PI);
    c.fillStyle = color || '#FFF';
    c.fill();
}

function drawLine(x1,y1,x2,y2,w,color){
    c.beginPath();
    c.lineWidth = w;
    c.strokeStyle = color;
    c.moveTo(x1,y1);
    c.lineTo(x2,y2);
    c.stroke();
}

function drawLightning(x1,y1,x2,y2){
    var x = x1,
        y = y1,
        nx,
        ny,
        distance,
        direction;

    drawCircle(c,x,y,3,'aqua');
    while(true){
        distance = randomInt(1,5);
        direction = getAngle(x,y,x2,y2) + randomFloat(-Math.PI/6,Math.PI/6);
        nx = x + distance * Math.cos(direction);
        ny = y + distance * Math.sin(direction);
        if(getDistance(nx ,ny ,x2 ,y2) > 5){
            drawLine(x, y, nx, ny, "5", "aqua");
            drawLine(x,y,nx,ny,'1','blue');
            x = nx;
            y = ny;
        }else{
            drawCircle(c,x,y,3,'aqua');
            drawLine(x, y, nx, ny, "5", "aqua");
            drawLine(x,y,nx,ny,'1','blue');
            break;
        }
    }
}
