//============================================= //
// Game.js
// Author: Jesse McCarthy
// Purpose: The game object of Inifinite Skies
//============================================= //

function Game(){


    //========================//
    //========== INIT ========//
    //========================//

	var	score      = 0,
		time       = 0,
		controls   = new Controls(),
		view       = new ViewPort( canvas.width/2, canvas.height/2 ),
        bg         = new BackgroundControl(),
        gui        = new GUI(),
		scoreBoard = new ScoreBoard(),
		collisions = new CollisionDetector(),
        player     = new Player(),
		objectGen  = new ObjectGenerator(),
		objs       = {
			enemies   : [ new Enemy( 100, 100 ,32 )],
			starDust  : [ new StarDust( 500, 500 )],
			orbs      : [],
			enemyOrbs : []
		};



    //==============================//
    //========== FUNCTIONS =========//
    //==============================//


    // *** UPDATE *** //

	this.background = function(){

		bg.update();
		drawBackground();

	}

    this.update = function(){

		incrementTime();
		updateObjects();
		draw();
		collisions.detect();
		garbageCollection();

		if( player.health <= 0 ) gameover();

    };

	function incrementTime(){

		if( ++time % fps === 0 ) score++;
		if( time % 150 == 0 && player.health < 100) player.health++;

	}

	function updateObjects(){

		player.update();
		iterateGameObjects( function( object ){
			object.update();
		});
		objectGen.update();
		view.update();
        bg.update();

	}

	function iterateGameObjects( callback ){

		for(var type in objs)
			objs[ type ].forEach( callback );

	}

	function garbageCollection(){

        iterateGameObjects(function(object, index, list){
			if( object.delete ){
				delete list[ index ];
				list.splice(index, 1);
			}
		});

    }

	function gameover(){

		state = 'post-game';
		getHighscoreTable( score );
		canvas.style.cursor = 'default';

	}


	// *** DRAW *** //

    function draw(){

        drawBackground();
        drawObjects();
        player.draw();
        gui.draw();
		scoreBoard.draw();

	};

	function drawBackground(){

		c.fillStyle = "#0a000d";
        c.fillRect(0,0,canvas.width,canvas.height);
        bg.draw();

	}

	function drawObjects(){

		iterateGameObjects(function( object ){
			object.draw();
		});

	}



    // ============================= //
    // ========== OBJECTS ========== //
    // ============================= //


	// *** Interfaces *** //

	function GameObjectInterface( x, y, size ){
		this.x = x;
		this.y = y;
		this.size   = size;
		this.delete = false;
		this.update = function(){};
		this.draw   = function(){};
	}

	function OrbInterface( x, y, direction ){

		GameObjectInterface.call(this,x,y,8);
		var orb     = this,
			speed   = 40,
			surplus = 400;

		this.update = function(){

			this.x += speed * Math.cos(direction);
			this.y += speed * Math.sin(direction);

			this.delete = !view.within( this.x, this.y, surplus );

		}

	}


	// *** Controls *** //

	function Controls(){

		this.up    = input.key.w;
		this.down  = input.key.s;
		this.left  = input.key.a;
		this.right = input.key.d;
		this.boost = input.key.space;

	}


	// *** Collisions *** //

    function CollisionDetector(){

        // all collision logic
		this.detect = function(){

			// player <-> enemy energy orb
			collisionBetween( player, objs.enemyOrbs, function( $1, $2 ){

				player.health -= 5;
				$2.destroy();

			});

			// player <-> star dust
			collisionBetween( player, objs.starDust, function( $1, $2 ){

				player.absorbStarDust();
				$2.destroy();
				sfx.index.starDust.play();

			});

			// enemy <-> energy orb
			collisionBetween( objs.enemies, objs.orbs, function( $1, $2 ){

				$1.destroy();
				$2.destroy();
				score += 10;

			});

		}

		function collisionBetween( objectList, objectList2, callback ){

			toArray(objectList).forEach(function(object, index, list){
				toArray(objectList2).forEach( function( object2, index2, list2){

					if( collision(object, object2) ){
						callback(
							new InstanceOfArray(object, index, list),
							new InstanceOfArray(object2, index2, list2)
						);
					}
				});
			});
		}

		function collision(object, object2){
			return (getDistance(object.x, object.y, object2.x, object2.y) < object.size + object2.size)
		}

		function toArray( object) {

			if( !Array.isArray(object) )
				object = [object];

			return object;

		}

		function InstanceOfArray(object, index, list){

			this.object = object;
			this.index  = index;
			this.list   = list;

			this.destroy = function(){
				object.delete = true
			}
		}

    }


	// *** Graphical User Interface *** //

	function ViewPort(){

		this.x = 0;
        this.y = 0;
	 	this.isShake = false;
		this.shakeCycle = new SineCycle( 5, 10 );


        this.update = function(){

			if(typeof player !== 'undefined'){

				while(player.x - this.x > canvas.width - canvas.width/2)   this.x++;
                while(player.x - this.x < canvas.width/2)                  this.x--;
                while(player.y - this.y > canvas.height - canvas.height/2) this.y++;
                while(player.y - this.y < canvas.height/2)                 this.y--;

			}

			shake();

		};

		this.shake = function(){

			this.isShake = true;

		}

		function shake(){

			if( this.isShake ){

				if( !shakeCycle.update().complete )
					this.x += shakeCycle.value;

				else{

					this.isShake = false;
					shakeCycle.reset();

				}
			}
		};

		this.within = function( x, y, surplus, size ){

			surplus = ( surplus != undefined ) ? surplus : 400;
			size    = size || 0;


			return(
				x > this.x - surplus - size / 2 &&
				x < this.x + canvas.width + surplus + size / 2 &&
				y > this.y - surplus - size / 2 &&
				y < this.y + canvas.height + surplus + size / 2
			);
		};

    }

    function GUI(){

        var imgHealth   = new Img(c, imgs.index.healthBar),
            imgPower    = new Img(c, imgs.index.powerBar),
            barWidth    = 118;

		this.draw = function(){

			drawHealthBar();
            drawEnergyBar();

		};

        function drawHealthBar(){

            imgHealth.draw({x : 10, y : 10});
            c.fillStyle = '#03b2ec';
            c.fillRect(15,15,barWidth * player.health / 100,19);

		}

        function drawEnergyBar(){

            imgPower.draw({ x : 10, y : 44 });
            c.fillStyle = '#8c47ff';
            c.fillRect( 15, 49, barWidth * player.energy / 100, 19 );

		}

    }

	function EnemyPointer(){

		this.sprite = new Img( c, imgs.index.arrow );

		this.draw = function(){

			this.sprite.draw({
				x: 100,
				y: 100,
				opacity : .5
			})

		}

	}

	function ScoreBoard(){

		var blur = new BlurCycle();

		this.draw = function(){

			blur.cycle();
			drawScore();

		}

		function drawScore(){

			c.save();

			c.font = 'italic 40px arcade';
			c.fillStyle = '#FFF';
			c.textAlign = 'center';

			c.shadowColor = 'aqua';
			c.shadowBlur = blur.value();

			c.fillText(score, canvas.width/2, 50);

			c.restore();

		}

		function BlurCycle(){

			var counter = 0,
				maxBlur = 20,
				blur;


			this.cycle = function(){

				var amplitude, offset;
				amplitude = offset = maxBlur/2;

				counter = (counter+=3) % 360;
				blur    = amplitude * Math.sin( toRadians( counter )) + 20;

			}

			this.value = function(){
				return blur;
			}

		}

	}

	function BackgroundControl(){


		// *** INIT *** //

		var timer     = new Timer([ 1, 3 ], createStar );
			bgStars   = [],
			starTypes = [
                imgs.index.aquaStar,
                imgs.index.purpleStar,
                imgs.index.whiteStar,
                imgs.index.blueStar
            ];

		var n = 0;
		while( n++ < 100 ) {
			timer.tick();
			updateStars();
			garbageCollection();
		}


		// *** METHODS *** //

		this.update = function(){

			timer.tick();
			updateStars();
			garbageCollection();

		};

		this.draw = function(){

			iterateStars( function( star ){
				star.draw();
			});

		};

		function iterateStars( callback ){

			for( var index = 0; index < bgStars.length; index++ )
				callback( bgStars[ index ], index );

		}

		function updateStars(){

			iterateStars( function( star ){
				star.update();
			});

		}

		function createStar(){

			var xmin = view.x - 200,
				xmax = view.x + canvas.width + 200,
				ymin = view.y - 200,
				ymax = view.y + canvas.height + 200;

			bgStars.push( new bgStar( randomInt( xmin, xmax ), randomInt( ymin, ymax )));

		}

		function garbageCollection(){

			iterateStars( function( star, index ){

				if( star.delete ) bgStars.splice( index, 1 );

			});
		}


		// *** OBJECTS *** //

		function bgStar(x,y){


			// *** INIT *** //

			var fader    = new Fader( 3 ),
				starType = starTypes[ randomInt( 0, starTypes.length - 1 )],
				scale    = randomFloat( .8, 1.4 ),
                rotate   = plusOrMinus( randomInt( 2, 5 )),
				opacity  = randomFloat( .3, .6 );

			this.sprite = new Img( c, starType );
			this.delete = false;
			this.x      = x;
            this.y      = y;


			// *** METHODS *** //

            this.update = function(){

				fader.fade();

				if( fader.complete ) this.delete = true;

			};

            this.draw = function(){
                this.sprite.draw({
                    x            : x - view.x,
                    y            : y - view.y,
                    scale        : scale * fader.opacity,
                    rotateSpd    : rotate,
                    centerOrigin : true,
					opacity      : opacity * fader.opacity
                });
            };


		}

	}


	// *** Game Objects *** //

    function Player(){


        // *** INIT *** //

        this.sprite = new Img(c, imgs.index.star);
        this.x      = canvas.width/2;
        this.y      = canvas.height/2;
        this.size   = 32;
        this.health = 100;
        this.energy  = 100;

        var player  = this,

			//image
			rotate  = 20,

			//movement
			speed   = 20,

			// boost
			boostModifier      = 0,
			boostState         = 'ready', // ready, active, cooldown
            boostCycle         = null,
			boostCooldownTimer = null,

			// cannon
			cannonState         = 'ready',
			cannonCooldownTimer = null;

        // *** METHODS *** //

		this.update = function(){

            movement();
			boost();
			cannon();

		};

		this.draw = function(){

			this.sprite.draw({

				x : this.x - view.x,
				y : this.y - view.y,
                w : this.size,
				h : this.size,

				rotateSpd    : rotate,
                centerOrigin : true,

				shadowColor : 'aqua',
				shadowBlur  : 60

			});

        };

		this.absorbStarDust = function(){

			if(( this.energy += 10 ) > 100 ) this.energy = 100;

		}


		// movement

        function movement(){

			var direction = getDirection(),
				speed     = getSpeed();

			commitMotion( direction, speed );

		};

		function getDirection(){

			var vector = {
				x : 0,
				y : 0
			};

			vector.x -= input.log[ controls.left ];
			vector.x += input.log[ controls.right ];
			vector.y -= input.log[ controls.up ];
			vector.y += input.log[ controls.down ];

			return getAngle( 0, 0, vector.x, vector.y );

		}

		function getSpeed(){

			return speed * getModifiers();

		}

		function getModifiers(){

			return ( 1 + boostModifier );

		}

		function commitMotion( direction, speed ){

			if( isNull( direction )) return;

			player.x += speed * Math.cos( direction );
			player.y += speed * Math.sin( direction );

		}


		// boost

		function boost(){

			if( boostState == 'ready' ){

				if( input.log[ controls.boost ] && hasSufficientEnergy( 10 )) activateBoost();

			}
			else if( boostState == 'active' ){

				boostModifier = boostCycle.update().value;

				if( boostCycle.complete ) completeBoost();

			}
			else if( boostState == 'cooldown' ){

				boostCooldownTimer.tick();

			}

		}

		function activateBoost(){

			applyEnergyCost( 10 );
			boostCycle = new SineCycle( 6, 3 );
			boostState = 'active';

		}

		function completeBoost(){

			boostCooldownTimer = new Timer( 30, cooldownBoost );
			boostState = 'cooldown';

		}

		function cooldownBoost(){

			boostState = 'ready';

		}


		// cannon

        function cannon(){

			if( cannonState == 'ready' ){

				if( input.mouse.left && hasSufficientEnergy( 1 )) shootCannon();

			}
			else if( cannonState == 'cooldown' ) cannonCooldownTimer.tick();

		};

		function shootCannon(){

			// shoot
			applyEnergyCost( 1 );
			objs.orbs.push( new EnergyOrb() );

			// begin cooldown
			cannonState = 'cooldown';
			cannonCooldownTimer = new Timer( 4, cooldownCannon );

			view.shake();

		}

		function cooldownCannon(){

			cannonState = 'ready';

		}


		// power

		function hasSufficientEnergy( cost ){

			return( player.energy >= cost );

		}

		function applyEnergyCost( cost ){

			return( player.energy -= cost);

		}

	}

    function Enemy( x, y, size ){

        // *** INIT *** //

		GameObjectInterface.call( this, x, y, size );
		this.sprite = new Img( c, imgs.index.enemy );

		var enemy       = this,
			active      = false,
			rotateSpeed = -26,

			// movement
			vx, vy, direction, distance, displacement,
			spd      = 20,
			waiting  = 0,
			inMotion = false,

			// fade in animation
			isFadingIn = true,
			fade       = new SineCycle( 60 );


        // *** METHODS *** //


	   // public

		this.update = function(){

			if( active ){
				movement();
			}
			else if( isFadingIn ){
				fadeIn();
			}

		};

		this.draw   = function(){

			this.sprite.draw({
				x : this.x - view.x,
				y : this.y - view.y,
				w : size,
				h : size,
				scale : fade.value,

				opacity      : fade.value / 2,
				rotateSpd    : rotateSpeed * fade.value,
				centerOrigin : true
			});

		};


		// fade in

		function fadeIn(){

			if( fade.vertex ){

				isFadingIn = false;
				active     = true;
				return;

			}

			fade.update();

		}

		// movement

		function movement(){

			if( inMotion ){

				if ( displacement < distance )
					continueMotion();
				else{
					resetMotion();
					shoot();
				}
			}
			else {
				if( waiting ) waitForMotion();
				else          beginMotion();
			}

		}

		function setDirection(){

			var ref = getAngle( enemy.x, enemy.y, player.x, player.y ),
				variance = Math.PI / 5,
				dir = ref + plusOrMinus(variance);

			return dir;

		}

		function setDistance(){

			var ref = getDistance( enemy.x, enemy.y, player.x, player.y );

			return ref + randomInt(50,150);

		};

		function beginMotion(){

			direction    = setDirection();
			distance     = setDistance();
			displacement = 0;
			vx           = spd * Math.cos( direction );
			vy           = spd * Math.sin( direction );
			inMotion     = true;

		}

		function continueMotion(){

			displacement += spd;
			enemy.x += vx;
			enemy.y += vy;

		}

		function resetMotion(){

			inMotion = false;
			waiting = randomInt( 15, 45 );

		}

		function waitForMotion(){

			if( --waiting === 0 ) inMotion = false;

		}


		// shoot

		function shoot(){

			objs.enemyOrbs.push( new EnemyEnergyOrb( enemy.x, enemy.y ));

		}

    }

    function EnergyOrb(){

		var parent    = this,
			x         = player.x,
			y         = player.y,
			direction = getDirection();

		OrbInterface.call( this, x, y, direction );

		function getDirection(){

			var rect = canvas.getBoundingClientRect(),
				mx   = input.mouse.x - rect.left,
				my   = input.mouse.y - rect.top;

			return getAngle( x-view.x, y-view.y, mx, my );

		}

        function drawElectricity(){

			var point  = randomOuterPoint(),
				point2 = randomOuterPoint();

            drawLightning( point.x, point.y, point2.x, point2.y );

		}

		function randomOuterPoint(){

			return {
				x : parent.x - view.x + parent.size * 1.1 * Math.cos(randomFloat(-Math.PI,Math.PI)),
				y : parent.y - view.y + parent.size * 1.1 * Math.sin(randomFloat(-Math.PI,Math.PI))
			}

		}

        this.draw = function(){

			drawCircle( c, this.x-view.x, this.y-view.y, this.size );
            drawElectricity();

		};

	}

	function EnemyEnergyOrb( x, y ){

		var	direction = getAngle(x, y, player.x, player.y);
		OrbInterface.call(this, x, y, direction);

		this.draw = function(){

			drawCircle(c, this.x - view.x, this.y - view.y, this.size, 'purple');

		};

	}

    function StarDust( x, y ){
        this.sprite = new Img(c,imgs.index.starDust);
        this.x = x;
        this.y = y;
        this.scale = .5;
        this.minScale = .5;
        this.maxScale = 1.25;
        this.grow = true;
        this.size = this.sprite.img[0].width * this.scale;

        this.update = function(){
            if(this.grow){
                this.scale += .05;
                if(this.scale >= this.maxScale)
                    this.grow = false;
            }
            else{
                this.scale -= .05;
                if(this.scale <= this.minScale)
                    this.grow = true;
            }
            this.size = this.sprite.img[0].width * this.scale;
        };

        this.draw = function(){
            this.sprite.draw({
                x : this.x - view.x,
                y : this.y - view.y,
                scale : this.scale,
                centerOrigin : true,
                rotateSpd : 26
            });
        };
    }


	// *** Object Generator *** //

	function ObjectGenerator(){

		var enemyGenerator    = new EnemyGenerator(),
			starDustGenerator = new StarDustGenerator();

		this.update = function(){

			enemyGenerator.update();
			starDustGenerator.update();

		};


		// *** OBJECTS *** //

		function EnemyGenerator(){

			var timer = new Timer([ 60, 90 ], generateEnemy );

			this.update = function(){

				timer.tick();

				if( objs.enemies.isEmpty() ){

					timer.reset();
					generateEnemy();

				}
			}

			function generateEnemy(){

				var x    = randomInt( view.x, view.x + canvas.width ),
					y    = randomInt( view.y, view.y + canvas.height ),
					size = randomInt( 28, 40 );

				objs.enemies.push( new Enemy( x, y, size ));

			}

		}

		function StarDustGenerator(){

			var timer     = new Timer([ 200, 400 ], createStarDust ),
				tolerance = 200; // min distance from other instances

			this.update = function(){

				timer.tick();

			}

			function createStarDust(){

				var coordinates = getSpawnCoordinates();

				objs.starDust.push( new StarDust( coordinates.x, coordinates.y ));

			}

			function getSpawnCoordinates(){

				var	x = randomInt( view.x, view.x + canvas.width ),
					y = randomInt( view.y, view.y + canvas.height );

				return({ x : x, y : y })

			}

		}
	}
}
